import copy
from typing import Union
from collections import defaultdict
from bson import ObjectId, DBRef
from edman.utils import Utils
from edman.exceptions import (EdmanInternalError, EdmanFormatError)
from edman import Config


class Convert:
    """
    研究データをEdman用にコンバートするクラス
    """

    def __init__(self) -> None:
        config = Config()  # システム環境用の設定を読み込む
        self.parent = config.parent
        self.child = config.child
        self.date = config.date

    def _get_child_reference(self, child_data: dict) -> dict:
        """
        子データのリファレンス情報を作成して取得

        :param dict child_data:
        :return:
        :rtype: dict
        """
        children = []
        for collection, child_value in child_data.items():

            # すでにparentが作られている場合は飛ばす
            if self.parent == collection:
                continue
            if isinstance(child_value, dict):
                children.append(DBRef(collection, child_value['_id']))
            elif isinstance(child_value, list) and (
                    not Utils.item_literal_check(child_value)):
                child_list = [DBRef(collection, j['_id']) for j in child_value]
                children.extend(child_list)
            else:
                continue

        return {self.child: children}

    def _convert_datetime(self, child_dict: dict) -> dict:
        """
        | 辞書内辞書になっている文字列日付時間データを、辞書内日付時間に変換

        | (例)
        | {'start_date': {'#date': '1981-04-23'}}
        | から
        | {'start_date': 1981-04-23T00:00:00}

        :param dict child_dict:
        :return: result
        :rtype: dict
        """
        result = copy.deepcopy(child_dict)
        if isinstance(child_dict, dict):
            try:
                for key, value in child_dict.items():
                    if isinstance(value, dict) and self.date in value:
                        result.update(
                            {
                                key: Utils.to_datetime(
                                    child_dict[key][self.date])
                            })
            except AttributeError:
                raise EdmanInternalError(
                    f'日付変換に失敗しました.構造に問題があります. {child_dict}')
        return result

    @staticmethod
    def _list_organize(extracted_data: list) -> list:
        """
        | リスト内辞書のデータから、同じコレクションの場合、
        | 値(リストになっている)を取り出し、マージさせる
        | バルクインサートを利用するために、
        | 同じコレクションでまとめたほうが効率が良いため
        | (例)
        | [
        |   {
        |        'honda': [
        |            {'name': 'NSX', 'power': '280'}
        |        ]
        |    },
        |    {
        |        'honda': [
        |            {'name': 'S800', 'first model year': '1966'}
        |        ]
        |    }
        | ]
        | から
        | [
        |    {
        |        'honda': [
        |            {'name': 'NSX', 'power': '280'},
        |            {'name': 'S800', 'first model year': '1966'}
        |        ]
        |    }
        | ]

        :param list extracted_data:
        :return: result
        :rtype: list
        """
        result = defaultdict(list)
        for insert_unit in extracted_data:
            collection = list(insert_unit.keys())[0]
            for doc in insert_unit[collection]:
                result[collection].append(doc)

        return [dict(result)]

    def _list_intercept_hook(self, collection: str,
                             doc_with_child: Union[dict, list]) -> dict:
        """
        | 対象ドキュメントの子要素のみを削除し、
        | 出力の対象コレクション内のリストに対して要素の追加もしくは書き換えを行う

        :param str collection:
        :param dict doc_with_child:
        :return: output
        :rtype: dict
        """

        def child_delete(doc: dict) -> None:
            """
            子要素を削除する

            :param dict doc:
            :return:
            """
            tmp = copy.deepcopy(doc)

            # 子要素のデータを抽出
            key_list = [k for k, v in doc.items()
                        if self.parent != k and self.child != k and (
                        isinstance(v, dict) or (
                        isinstance(v, list) and (
                        not Utils.item_literal_check(v))))]

            # 該当データがtmpにあれば削除
            tmp = {k: v for k, v in tmp.items() if k not in key_list}

            # outputのデータを入れ替える
            if collection in output:
                output[collection].append(tmp)
            else:
                output[collection] = [tmp]

        output = {}
        if isinstance(doc_with_child, list):
            for i in doc_with_child:
                child_delete(i)
        else:
            child_delete(doc_with_child)

        return output

    def _ref(self, raw_data: dict) -> list:
        """
        リファレンスモードでedman用に変換を行う

        :param dict raw_data:
        :return:
        :rtype: list
        """
        list_output = []
        ref_list = []

        def recursive(reading_dict_data: dict) -> dict:
            """
            edman用に変換を行う
            再帰
            要リファクタリング

            :param dict reading_dict_data:
            :return: output
            :rtype: dict
            """
            output = {}
            parent = -2  # 説明変数
            my = -1  # 説明変数

            for key, value in reading_dict_data.items():

                if isinstance(value, dict):

                    if not Utils.collection_name_check(key):
                        raise EdmanFormatError(f'この名前はコレクション名にできません {key}')

                    # リファレンス作成
                    ref_list.append(DBRef(key, ObjectId()))

                    # tmpから子データが返ってくる
                    tmp = recursive(self._convert_datetime(value))

                    # 親のリファレンス(コレクション)を追加
                    # rootの場合は追加されない
                    if len(ref_list) > 1:
                        tmp.update({self.parent: ref_list[parent]})

                    # 子データのリファレンスを取得して親のデータに入れる
                    child_ref = self._get_child_reference(tmp)
                    if list(child_ref.values())[0]:  # 子データがない場合もある
                        tmp.update(child_ref)

                    # self.parentがない場合(=root)の場合にoidを追加する
                    if self.parent not in tmp:
                        tmp.update({'_id': ref_list[0].id})

                    del ref_list[my]

                    # バルクインサート用のリストを作成
                    list_output.append(self._list_intercept_hook(key, tmp))

                    output.update({key: tmp})

                elif isinstance(value, list) and Utils.item_literal_check(
                        value):
                    if not Utils.field_name_check(key):
                        raise EdmanFormatError(f'フィールド名に不備があります {key}')

                    # 日付データが含まれていたらdatetimeオブジェクトに変換
                    output.update({key: self._date_replace(value)})

                elif isinstance(value, list):

                    if not Utils.collection_name_check(key):
                        raise EdmanFormatError(f'この名前はコレクション名にできません {key}')

                    tmp_list = []
                    for i in value:
                        # リファレンス作成
                        ref_list.append(DBRef(key, ObjectId()))

                        # tmpから子データが返ってくる
                        tmp = recursive(self._convert_datetime(i))

                        # 親のリファレンス(コレクション)を追加
                        # rootの場合は追加されない
                        if len(ref_list) > 1:
                            tmp.update({
                                self.parent: ref_list[parent]
                            })

                        # 子要素だけのデータの場合elseへ飛ばないのでここでoidを作成する
                        if '_id' not in tmp:
                            tmp.update({'_id': ref_list[my].id})

                        # 子データのリファレンスを取得して親のデータに入れる
                        child_ref = self._get_child_reference(tmp)
                        if list(child_ref.values())[0]:  # 子データがない場合もある
                            tmp.update(child_ref)

                        del ref_list[my]
                        tmp_list.append(tmp)

                    # バルクインサート用のリストを作成
                    list_output.append(
                        self._list_intercept_hook(key, tmp_list))

                    output.update({key: tmp_list})

                else:
                    if not Utils.field_name_check(key):
                        raise EdmanFormatError(f'フィールド名に不備があります {key}')

                    # oidを取得して追加(rootの場合、すでに作成されている場合がある？)
                    if '_id' not in output:
                        output.update({'_id': ref_list[my].id})
                    output.update({key: value})

            return output

        # list_outputを書き換えているため、extract()の返り値(output)は利用していない
        _ = recursive(raw_data)
        return self._list_organize(list_output)

    @staticmethod
    def pullout_key(data: dict, pull_key: str) -> dict:
        """
        最初に発見した指定のキーの要素を、子要素を含めて抜き出す

        :param dict data:
        :param str pull_key:
        :return: output
        :rtype: dict
        """

        def recursive(doc):
            for key, value in doc.items():

                # 最初に発見した要素がoutputに入っていたら再帰を終了
                if output:
                    break
                if isinstance(value, dict):
                    if key == pull_key:
                        output.update({key: value})
                        break
                    recursive(value)
                elif isinstance(value, list) and Utils.item_literal_check(
                        value):
                    continue
                elif isinstance(value, list):
                    if key == pull_key:
                        output.update({key: value})
                        break
                    [recursive(i) for i in value]
                else:
                    continue

        output = {}
        recursive(data)
        return output

    @staticmethod
    def exclusion_key(data: dict, ex_keys: tuple) -> dict:
        """
        指定のキーの要素を除外する

        :param dict data:
        :param tuple ex_keys:
        :return:
        :rtype: dict
        """

        def recursive(doc):
            output = {}
            for key, value in doc.items():
                if isinstance(value, dict):
                    if key in ex_keys:
                        continue
                    o = {key: recursive(value)}
                elif isinstance(value, list) and Utils.item_literal_check(
                        value):
                    o = {key: value}
                elif isinstance(value, list):
                    if key in ex_keys:
                        continue
                    o = {key: [recursive(i) for i in value]}
                else:
                    o = {key: value}
                output.update(o)
            return output

        return recursive(data)

    @staticmethod
    def _attached_oid(data: dict) -> dict:
        """
        | 辞書の一番上の階層にoidを付与する
        | 辞書の中身がリストの場合はリスト内の辞書すべてにoidを付与

        :param dict data:
        :return: data
        :rtype: dict
        """
        for k, v in data.items():
            try:
                if isinstance(v, dict):
                    v.update({'_id': ObjectId()})
                else:
                    for doc in v:
                        doc.update({'_id': ObjectId()})
            except ValueError:
                raise
        return data

    def _date_replace(self, list_data: list) -> list:
        """
        | リスト内の要素に{'#date':日付時間}のデータが含まれていたら
        | datetimeオブジェクトに変換する
        | 例
        | [{'#date':2019-02-28 11:43:22}, ' test_date']
        | ↓
        | [datetime.datetime(2019, 2, 28, 11, 43, 22), 'test_date']

        :param list list_data:
        :return:
        :rtype: list
        """
        return [Utils.to_datetime(i[self.date])
                if isinstance(i, dict) and self.date in i
                else i
                for i in list_data]

    def emb(self, raw_data: dict) -> dict:
        """
        | エンベデッドモードでedman用の変換を行う
        | 主に日付の変換
        | update処理でも使用している

        :param dict raw_data:
        :return:
        :rtype: dict
        """

        def recursive(data: dict) -> dict:
            """
            再帰で辞書を走査して、日付データの変換などを行う
            要リファクタリング

            :param dict data:
            :return:
            :rtype: dict
            """
            output = {}
            for key, value in data.items():
                if isinstance(value, dict):
                    if not Utils.collection_name_check(key):
                        raise EdmanFormatError(f'この名前は使用できません {key}')
                    converted_value = self._convert_datetime(value)
                    o = {key: recursive(converted_value)}

                elif isinstance(value, list):
                    # 日付データが含まれていたらdatetimeオブジェクトに変換
                    value = self._date_replace(value)
                    # 通常のリストデータの場合
                    if Utils.item_literal_check(value):
                        if not Utils.field_name_check(key):
                            raise EdmanFormatError(f'フィールド名に不備があります {key}')
                        list_tmp_data = value
                    # 子要素としてのリストデータの場合
                    else:
                        if not Utils.collection_name_check(key):
                            raise EdmanFormatError(f'この名前は使用できません {key}')
                        list_tmp_data = [recursive(self._convert_datetime(i))
                                         for i in value]
                    o = {key: list_tmp_data}
                else:
                    if not Utils.field_name_check(key):
                        raise EdmanFormatError(f'フィールド名に不備があります {key}')
                    o = {key: value}
                output.update(o)
            return output

        result = recursive(raw_data)
        return result

    def dict_to_edman(self, raw_data: dict, mode='ref') -> list:
        """
        | json辞書からedman用に変換する
        | embはobjectIdを付与したり、辞書からリストに変換している

        :param dict raw_data: JSONを辞書にしたデータ
        :param str mode: ref(reference) or emb(embedded) データ構造の選択肢
        :return: インサート用のリストデータ
        :rtype: list
        """
        if mode == 'ref':
            return self._ref(raw_data)
        elif mode == 'emb':
            return [self._attached_oid(self.emb(raw_data))]
        else:
            raise EdmanFormatError("投入モードは'ref'または'emb'です")



class Config:
    """
    このパッケージを利用したシステム上での共通の定義

    **デフォルトのままをおすすめする**

      | 定義を変更した場合、このパッケージを利用している他のシステムと、データ交換ができなくなる
      | DBにデータが入っている状態で、この定義を変更した場合、データが破壊される可能性が高い

    | それでも変更したい場合は、単一、もしくは同じシステム内で定義を統一すること
    | その場合、他のシステムとデータ交換したくなった場合は独自に変換スクリプトを作成すること
    """
    parent = '_ed_parent'  # 親のリファレンス情報
    child = '_ed_child'  # 子のリファレンス情報
    file = '_ed_file'  # Grid.fsのリファレンス情報
    date = '#date'  # 日付に変換する場合のタグ

    # Gzip 圧縮レベル (ファイル圧縮時に使用)
    gzip_compress_level = 6

    # 再帰で回数の上限に達した場合、下記を適切な値で設定
    # import sys
    # sys.setrecursionlimit(1000)  # pythonのデフォルトは1000



import sys
import os
import copy
import gzip
from typing import Union, Tuple, Iterator, List
from pathlib import Path
import gridfs
import jmespath
from bson import ObjectId
from edman.utils import Utils
from edman.exceptions import (EdmanFormatError, EdmanDbProcessError)
from edman import Config


class File:
    """
    ファイル取扱クラス
    """

    def __init__(self, db=None) -> None:

        if db is not None:
            self.db = db
            self.fs = gridfs.GridFS(self.db)
        self.file_ref = Config.file
        self.comp_level = Config.gzip_compress_level

    @staticmethod
    def file_gen(files: Tuple[Path]) -> Iterator:
        """
        ファイルタプルからファイルを取り出すジェネレータ

        :param tuple files: 中身はPathオブジェクト
        :return: ファイル名と内容(str)のタプル
        :rtype: tuple
        """
        for file in files:
            try:
                with file.open('rb') as f:
                    fp = f.read()
            except IOError:
                raise
            yield file.name, fp

    def add_file_reference(self, collection: str, oid: Union[ObjectId, str],
                           file_path: Tuple[Path], structure: str,
                           query=None, compress=False) -> bool:
        """
        ドキュメントにファイルリファレンスを追加する
        ファイルのインサート処理、圧縮処理なども行う

        :param str collection:
        :param oid:
        :type oid: ObjectId or str
        :param tuple file_path:
        :param str structure:
        :param query:
        :type query: list or None
        :param bool compress: default False
        :return:
        :rtype: bool
        """

        oid = Utils.conv_objectid(oid)

        # ドキュメント存在確認&対象ドキュメント取得
        doc = self.db[collection].find_one({'_id': oid})
        if doc is None:
            # TODO
            sys.exit('対象のドキュメントが存在しません')

        if structure == 'emb':
            # クエリーがドキュメントのキーとして存在するかチェック
            if not Utils.query_check(query, doc):
                raise EdmanFormatError('対象のドキュメントに対してクエリーが一致しません.')

        # ファイルのインサート
        inserted_file_oids = []
        for file in self.file_gen(file_path):
            file_obj = file[1]
            metadata = {'filename': file[0]}

            if compress:
                file_obj = gzip.compress(file_obj,
                                         compresslevel=self.comp_level)
                metadata.update({'compress': 'gzip'})

            inserted_file_oids.append(self.fs.put(file_obj, **metadata))

        if structure == 'ref':
            new_doc = self.file_list_attachment(doc, inserted_file_oids)

        elif structure == 'emb':
            try:
                new_doc = Utils.doc_traverse(doc, inserted_file_oids, query,
                                             self.file_list_attachment)
            except Exception:
                raise
        else:
            raise EdmanFormatError('構造はrefかembが必要です')

        # ドキュメント差し替え
        replace_result = self.db[collection].replace_one({'_id': oid}, new_doc)

        if replace_result.modified_count == 1:
            return True
        else:  # 差し替えができなかった時は添付ファイルは削除
            self.fs_delete(inserted_file_oids)
            return False

    def delete(self, delete_oid: ObjectId, collection: str,
               oid: Union[ObjectId, str], structure: str, query=None) -> bool:
        """
        該当のoidをファイルリファレンスから削除し、GridFSからファイルを削除

        :param ObjectId delete_oid:
        :param str collection:
        :param str oid:
        :param str structure:
        :param query:
        :type query: list or None
        :return:
        :rtype: bool
        """
        oid = Utils.conv_objectid(oid)

        # ドキュメント存在確認&コレクション存在確認&対象ドキュメント取得
        doc = self.db[collection].find_one({'_id': oid})
        if doc is None:
            # TODO
            sys.exit('対象のコレクション、またはドキュメントが存在しません')

        # ファイルリスト取得
        files_list = self.get_file_ref(doc, structure, query)

        # リファレンスデータを編集
        if len(files_list) > 0:
            # 何らかの原因で重複があった場合を避けるため一度setにする
            files_list = list(set(files_list))
            files_list.remove(delete_oid)
        else:
            raise EdmanDbProcessError('ファイルが存在しません')

        # ドキュメントを新しいファイルリファレンスに置き換える
        if structure == 'ref':
            try:
                new_doc = self.file_list_replace(doc, files_list)
            except Exception:
                raise

        elif structure == 'emb':
            try:
                new_doc = Utils.doc_traverse(doc, files_list, query,
                                             self.file_list_replace)
            except Exception:
                raise
        else:
            raise EdmanFormatError('structureはrefまたはembの指定が必要です')

        replace_result = self.db[collection].replace_one({'_id': oid}, new_doc)

        # fsから該当ファイルを削除
        if replace_result.modified_count:
            self.fs_delete([delete_oid])

        # ファイルが削除されたか検証
        if self.fs.exists(delete_oid):
            return False
        else:
            return True

    def fs_delete(self, oids: list) -> None:
        """
        fsからファイル削除

        :param list oids:
        :return:
        """
        if len(oids):
            for oid in oids:
                if self.fs.exists(oid):
                    self.fs.delete(oid)

    def get_file_ref(self, doc: dict, structure: str, query=None) -> list:
        """
        ファイルリファレンス情報を取得

        :param dict doc:
        :param str structure:
        :param query:
        :type: list or None
        :return: files_list
        :rtype: list
        """
        if structure == 'emb' and query is None:
            raise EdmanFormatError('embにはクエリが必要です')
        if structure != 'emb' and structure != 'ref':
            raise EdmanFormatError('構造の選択はembまたはrefが必要です')

        files_list = []
        if structure == 'ref':
            if self.file_ref in doc:
                files_list = doc[self.file_ref]
        else:
            if not Utils.query_check(query, doc):
                EdmanFormatError('対象のドキュメントに対してクエリーが一致しません.')
            # docから対象クエリを利用してファイルのリストを取得
            # deepcopyを使用しないとなぜか子のスコープのqueryがクリヤーされる
            query_c = copy.deepcopy(query)
            try:
                files_list = self._get_emb_files_list(doc, query_c)
            except Exception:
                raise

        return files_list

    def get_file_names(self, collection: str, oid: Union[ObjectId, str],
                       structure: str, query=None) -> dict:
        """
        ファイル一覧を取得
        ファイルが存在しなければ空の辞書を返す

        :param str collection:
        :param str oid:
        :param str structure:
        :param query: embの時だけ必要. refの時はNone
        :type query: list or None
        :return: result
        :rtype: dict
        """
        oid = Utils.conv_objectid(oid)
        result = {}

        # ドキュメント存在確認&コレクション存在確認&対象ドキュメント取得
        doc = self.db[collection].find_one({'_id': oid})
        if doc is None:
            raise EdmanDbProcessError(
                f'ドキュメントまたはコレクションが存在しません oid:{oid} collection:{collection}')

        # gridfsからファイル名を取り出す
        for file_oid in self.get_file_ref(doc, structure, query):
            try:
                fs_out = self.fs.get(file_oid)
            except gridfs.errors.NoFile:
                pass
            else:
                result.update({file_oid: fs_out.filename})
        return result

    def download(self, oid: ObjectId, path: Union[str, Path]) -> bool:
        """
        Gridfsからデータをダウンロードし、ファイルに保存
        metadataに圧縮指定があれば伸長する

        :param ObjectId oid:
        :param path:
        :type path: str or Path
        :return: result
        :rtype: bool
        """
        result = False

        # パスがstrならpathlibにする
        p = Path(path) if isinstance(path, str) else path

        # パスが正しいか検証
        if not p.exists():
            raise FileNotFoundError

        # ダウンロード処理
        if self.fs.exists(oid):
            fs_out = self.fs.get(oid)
            save_path = p / fs_out.filename
            try:
                with save_path.open('wb') as f:
                    tmp = fs_out.read()
                    if hasattr(fs_out,
                               'compress') and fs_out.compress == 'gzip':
                        tmp = gzip.decompress(tmp)
                    f.write(tmp)
                    f.flush()
                    os.fsync(f.fileno())
            except IOError:
                raise

            if save_path.exists():
                result = True

        else:
            raise EdmanDbProcessError('指定のファイルはDBに存在しません')

        return result

    def file_list_attachment(self, doc: dict,
                             files_oid: List[ObjectId]) -> dict:
        """
        辞書データにファイルのoidを挿入する
        docにself.file_refがあれば、追加する処理
        oidが重複していないものだけ追加
        ファイルが同じでも別のoidが与えられていれば追加される

        :param dict doc:
        :param list files_oid: ObjectIdのリスト
        :return: doc
        :rtype: dict
        """
        if self.file_ref in doc:
            doc[self.file_ref].extend(files_oid)
            files_oid = sorted(list(set(doc[self.file_ref])))
        # self.file_refがなければ作成してfiles_oidを値として更新
        if len(files_oid):
            doc.update({self.file_ref: files_oid})
        return doc

    def file_list_replace(self, doc: dict, files_oid: list) -> dict:
        """
        ドキュメントのファイルリファレンスを入力されたリストに置き換える
        もし空リストならファイルリファレンス自体を削除する
        すでにファイルリファレンスデータが存在していることを前提としているため、
        docにファイルリファレンスデータが無かった場合は例外を発生する

        :param dict doc:
        :param list files_oid:
        :return: doc
        :rtype: dict
        """
        if self.file_ref in doc:
            if len(files_oid):
                doc[self.file_ref] = files_oid
            else:
                del doc[self.file_ref]
        else:
            raise ValueError(f'{self.file_ref}がないか削除された可能性があります')
        return doc

    def _get_emb_files_list(self, doc: dict, query: list) -> list:
        """
        ドキュメントからファイルリファレンスのリストを取得する

        :param dict doc:
        :param list query:
        :return:
        :rtype:list
        """
        s = Utils.generate_jms_query(query)
        s += '.' + self.file_ref + '[]'
        return jmespath.search(s, doc)



from datetime import datetime
from typing import Union
from pymongo import errors
from bson import errors, ObjectId
from edman.utils import Utils
from edman.exceptions import EdmanDbProcessError
from edman import Config


class Search:
    """
    検索関連クラス

    """

    def __init__(self, db=None) -> None:
        config = Config()  # システム環境用の設定を読み込む
        self.parent = config.parent
        self.child = config.child
        self.date = config.date
        self.file = config.file
        self.db = db

        if self.db is not None:
            self.connected_db = db.get_db

    def find(self, collection: str, query: dict, parent_depth: int,
             child_depth: int, exclusion=None) -> dict:
        """
        検索用メソッド

        :param str collection: 対象コレクション
        :param dict query: 検索クエリ
        :param int parent_depth: 親の指定深度
        :param int child_depth: 子の指定深度
        :param None or list exclusion:除外するリファレンスキー 例 ['_ed_file']
        :return: result 親 + 自分 + 子の階層構造となった辞書データ
        :rtype: dict
        """

        coll_filter = {"name": {"$regex": r"^(?!system\.)"}}
        if collection not in self.connected_db.list_collection_names(
                filter=coll_filter):
            raise EdmanDbProcessError('コレクションが存在しません')

        query = self._objectid_replacement(query)
        self_result = self._get_self(query, collection)
        if self_result is None:
            raise EdmanDbProcessError('データを取得できませんでした')
        reference_point_result = self.db.get_reference_point(
            self_result[collection])

        parent_result = None
        if reference_point_result[self.parent]:
            parent_result = self._get_parent(self_result, parent_depth)

        children_result = None
        if reference_point_result[self.child]:
            children_result = self.db.get_child(self_result, child_depth)

        # 親も子も存在しない時はselfのみ
        result = self_result

        # 子データが存在する時だけselfとマージ
        if children_result:
            self_key = list(self_result.keys())[0]
            self_result[self_key].update(children_result)

        # 親データが存在する時だけselfとマージ
        if parent_result:
            result = self._merge_parent(parent_result, result)

        # JSONデータ用に変換
        result = self.process_data_derived_from_mongodb(result,
                                                        exclusion=exclusion)

        return result

    def _merge_parent(self, parent_result: dict, self_result: dict) -> dict:
        """
        親データに家族データをマージする

        :param dict parent_result:マージ前の親データ
        :param dict self_result:マージ済み家族データ
        :return: parent_result 親と自分と子をマージしたデータ
        :rtype: dict
        """

        def recursive(data: dict):
            """
            親データ内で一番深いドキュメントに家族データを挿入する

            :param dict data:
            :return:
            """
            for key in data.keys():
                # 親が複数存在することはありえないので辞書のみで可
                if isinstance(data[key], dict):
                    if '_id' in data[key]:
                        if parent_id == data[key]['_id']:
                            data[key].update(self_result)
                    recursive(data[key])

        self_doc = list(self_result.values())[0]
        parent_id = self_doc[self.parent].id
        recursive(parent_result)  # parent_resultにself_resultを入れる
        return parent_result

    @staticmethod
    def _objectid_replacement(query: dict) -> dict:
        """
        ObjectIdのチェックと変換

        :param dict query:
        :return: query
        :rtype: dict
        """
        if '_id' in query:
            try:
                query['_id'] = ObjectId(query['_id'])
            except errors.InvalidId:
                raise
        return query

    def _get_self(self, query: dict, collection: str) -> Union[dict, None]:
        """
        自分自身のドキュメント取得

        :param dict query:
        :param str collection:
        :return:
        :rtype: dict or None
        """
        result = None
        try:
            docs = list(self.connected_db[collection].find(query))
        except errors.OperationFailure:
            raise EdmanDbProcessError('ドキュメントが取得できませんでした')
        else:
            if len(docs):
                # 複数ドキュメントの場合は選択処理へ
                # doc = docs[0] if len(docs) == 1 else self._self_data_select(docs)
                # result = {collection: docs[0]}
                result = {collection: docs[0]}
        return result

    def _get_parent(self, self_doc: dict, depth: int) -> Union[dict, None]:
        """
        | 親となるドキュメントを取得
        | depthで深度を設定し、階層分取得する

        :param dict self_doc:
        :param int depth:
        :return: result
        :rtype: dict or None
        """

        def recursive(doc):
            """
            再帰でドキュメントを取得
            DBReferenceを利用し、設定されている深度を減らしながら再帰
            """
            if self.parent in doc:
                parent = self.connected_db.dereference(doc[self.parent])
                parent_collection = doc[self.parent].collection
                result.append({parent_collection: parent})
                nonlocal depth
                depth -= 1
                if depth > 0:
                    recursive(parent)

        if depth > 0:
            result = []  # recによって書き換えられる
            recursive(list(self_doc.values())[0])
            result = self._build_to_doc_parent(result)
        else:
            result = None
        return result

    @staticmethod
    def _build_to_doc_parent(parent_data_list: list) -> dict:
        """
        | 親の検索結果（リスト）を入れ子辞書に組み立てる
        |
        | 入力値であるparent_data_listについて、
        | parentに近い方から順番に並んでいる(一番最後がrootまたはrootに近い方)

        :param list parent_data_list:
        :return: result
        :rtype: dict
        """
        result = None
        for read_data in parent_data_list:
            if result is not None:
                buff = read_data
                collection = list(buff.keys())[0]
                doc = list(buff.values())[0]
                doc.update(result)
                result = {collection: doc}
            else:
                result = read_data
        return result

    def process_data_derived_from_mongodb(self, result_dict: dict,
                                          exclusion=None) -> dict:
        """
        MongoDB依存の項目を処理する::
          _idとrefの削除
          型をJSONに合わせる

        :param dict result_dict:
        :param List or None exclusion:
            default ['_id', self.parent, self.child, self.file]
        :return: result_dict
        :rtype: dict
        """
        default_refs = ['_id', self.parent, self.child, self.file]
        if exclusion is None:
            refs = tuple(default_refs)
        else:
            if not isinstance(exclusion, list):
                raise ValueError('listもしくはNoneが必要です')
            else:
                for i in exclusion:
                    if i not in default_refs:
                        raise ValueError(
                            f"{['_id', self.parent, self.child, self.file]}"
                            'の中から選択する必要があります')
            # デフォルトの値からexclusionを差し引く
            refs = tuple(set(default_refs) - set(exclusion))

        def recursive(data: dict):
            # idとrefの削除
            for key, val in data.items():
                if isinstance(data[key], dict):
                    recursive(Utils.item_delete(data[key], refs))
                # リストデータは中身を型変換する
                elif isinstance(data[key], list) and Utils.item_literal_check(
                        data[key]):
                    data[key] = [self._format_datetime(i) for i in data[key]]
                elif isinstance(data[key], list):
                    for item in data[key]:
                        recursive(Utils.item_delete(item, refs))
                else:
                    try:  # 型変換
                        data[key] = self._format_datetime(data[key])
                    except Exception:
                        raise

        recursive(result_dict)
        return result_dict

    def _format_datetime(self, item: Union[str, int, float, bool, datetime]
                         ) -> Union[dict, str, int, float, bool, datetime]:
        """
        datetime型なら書式変更して辞書に入れる
        その場合は辞書を返し、それ以外の時は入ってきた値をそのまま返す

        :param item:
        :type item: str or int or float or bool or datetime
        :return: result
        :rtype: dict or str or int or float or bool or datetime
        """
        # datetimeそのままだと%Y-%m-%dと%H:%M:%Sの間に"T"が入るため書式変更
        result = item
        if isinstance(item, datetime):
            result = {self.date: item.strftime("%Y-%m-%d %H:%M:%S")}
        return result



import os
from datetime import datetime
from pathlib import Path
from typing import Union
from bson.json_util import dumps
from edman.exceptions import EdmanFormatError

class JsonManager:
    """
    JSONファイルの取扱いクラス
    """

    @staticmethod
    def save(report_data: dict, path: Union[str, Path], name: str,
             date=False) -> None:
        """
        JSONファイルに書き出し

        :param dict report_data: 対象の辞書データ
        :param path: ファイルパス
        :type path: str or Path
        :param str name: ファイル名
        :param bool date: 日付 ファイル名先頭に追加
        :return: None
        """
        if not isinstance(report_data, dict):
            raise EdmanFormatError('Not Dict Data')

        date_str = datetime.today().strftime("%Y%m%d%H%M%S%f") + "_" if date else ""
        filename = date_str + name + '.json'
        p = path if isinstance(path, Path) else Path(path)
        savepath = p / filename

        with savepath.open("w", encoding='utf8') as file:
            # TODO try except
            file.write(dumps(report_data, ensure_ascii=False, indent=4))
            file.flush()
            os.fsync(file.fileno())



class EdmanError(Exception):
    def __init__(self, message):
        self._message = message

    def __str__(self):
        return f'{self.__class__.__name__} {self._message}'


class EdmanDbConnectError(EdmanError):
    """
    DB接続に関するエラー
    """
    pass


class EdmanDbProcessError(EdmanError):
    """
    DB実行処理に関するエラー
    """
    pass


class EdmanInternalError(EdmanError):
    """
    Edmanの処理に関するエラー
    """
    pass


class EdmanFormatError(EdmanError):
    """
    Edmanの名称や決まりごとに関するエラー
    """
    pass
